#include <mes2.h>
#include <string.h>

#define	TFTP_RETRY	4

#define TFTP_SIZE	512

#define	TFTP_NOFILE	0
#define	TFTP_ACCESS	2
#define	TFTP_ILLEGAL	4

#define	TFTP_READREQ	1
#define	TFTP_WRITEREQ	2
#define	TFTP_DATA	3
#define TFTP_ACK	4
#define TFTP_ERROR	5

#define	FILE_RETRY	8

static int make_req_packet(char *packet, char *name, char command) {
	int	index;

	packet[0] = 0;
	packet[1] = command;
	index = 2;
	strcpy(&packet[index], name);
	index += strlen(name) + 1;
	strcpy(&packet[index], "octet");
	index += strlen("octet") + 1;
	return index;
}

static int ack_get(int sock,struct sockaddr *target) {
	unsigned char	ack[4];
	int		ret, num;

	ret = recvfrom(sock, ack, 4, target);
	if(ret == -1 || ack[1] != TFTP_ACK) return -1;
	num = (ack[2] << 8) & 0xff00;
	num += ack[3] & 0xff;
	return num;
}

static void ack_put(int sock,struct sockaddr *target, unsigned short pnum) {
	unsigned char	ack[4];

	ack[0] = 0;
	ack[1] = TFTP_ACK;
	ack[2] = pnum >> 8;
	ack[3] = pnum;
	sendto(sock, (char*)ack, 4, target);
}

static int tftp_common(int ip, char *filename, char command, void (*func)(), int timeout) {
	int		ack, i, c, sock, size, filesize, count, siz;
	int		fd, mode, fullsize, ret, retry, f_retry;
	unsigned int	myip, servip, mask;
	unsigned char	mac[6];
	unsigned char	*packet;
	unsigned char	resp[4];
	unsigned short	pnum, rnum;
	struct sockaddr	target;

	switch(command) {
	case TFTP_WRITEREQ:
		mode = OptRead;
		fullsize = TFTP_SIZE;
		break;
	case TFTP_READREQ:
		mode = OptWrite;
		fullsize = TFTP_SIZE + 4;
		break;
	default:
		return -1;
	}
	fd = open(filename, mode);
	if(fd == -1) return -1;
	servip = ip;
	packet = malloc(TFTP_SIZE + 4);
	if(packet == 0) {
		close(fd);
		return -1;
	}
	sock = udp_socket();
	if(sock == -1) {
		close(fd);
		free(packet);
		return -1;
	}
	udp_timeout(sock, timeout * 1000);
	bzero(packet, TFTP_SIZE + 4);
	target.sin_addr = servip;
	target.sin_port = 69;
        udp_bind(sock, &target);
	size = make_req_packet(packet, filename, command);
	sendto(sock, packet, size, &target);
	ret = -1;
	count = 0;
        if(command == TFTP_READREQ || ack_get(sock, &target) == 0) {
        	ret = 0;
	        pnum = 1;
		do {
			bzero(packet, TFTP_SIZE + 4);
			switch(command) {
			case TFTP_WRITEREQ:
				size = read(fd, &packet[4], TFTP_SIZE);
				if(size == -1) {
					ret = -1;
					break;
				}
				for(retry = 0;retry < TFTP_RETRY;retry++) {
					packet[0] = 0;
					packet[1] = TFTP_DATA;
					packet[2] = pnum >> 8;
					packet[3] = pnum;
					sendto(sock, packet, size + 4, &target);
					ack = ack_get(sock, &target);
					if(ack == -1) ret = -1;
					if(ack == pnum || ack == -1) break;
				}
				if(retry == TFTP_RETRY) ret = -1;
				break;
			case TFTP_READREQ:
				for(retry = 0;retry < TFTP_RETRY;retry++) {
					size = recvfrom(sock, packet, TFTP_SIZE + 4, &target);
					rnum = (packet[2] << 8) & 0xff00;
					rnum += packet[3] & 0xff;
					if(size >= 0 && pnum == rnum) {
						for(f_retry = 0;f_retry < FILE_RETRY;f_retry++) {
							siz = write(fd, &packet[4], size - 4);
							if(siz != -1) break;
							close(fd);
							fd = open(filename, OptAppend);
							if(fd == -1) {
								f_retry = FILE_RETRY;
								break;
							}
						}
						if(f_retry == FILE_RETRY) {
							ret = -1;
							break;
						}
						resp[0] = 0;
						resp[1] = TFTP_ACK;
						resp[2] = packet[2];
						resp[3] = packet[3];
						sendto(sock, (char*)resp, 4, &target);
						break;
					} else {
						sendto(sock, (char*)resp, 4, &target);
					}
				}
				if(retry == TFTP_RETRY) ret = -1;
				break;
			}
			pnum++;
			if((count++ % 10) == 0) if(func != 0) (*func)();
		} while(size == fullsize && ret == 0);
	}
	close(fd);
	udp_free(sock);
	free(packet);
	return ret;
}

int tftp_save(int ip, char *filename, void (*func)(), int timeout) {
	return tftp_common(ip, filename, TFTP_WRITEREQ, func, timeout);
}

int tftp_load(int ip, char *filename, void (*func)(), int timeout) {
	return tftp_common(ip, filename, TFTP_READREQ, func, timeout);
}

